#!/usr/bin/env python
# Time-stamp: <Dotplot.py 2015-05-19 12:46:50 Mark Voorhies>
"""
pydotter, a Python clone of Erik Sonnhammer's DOTTER

Based on the algorithms described in Gene 167:GC1 (Sonnhammer and Durbin, 1996)
"""

from math import sqrt, ceil

from FastaFile import complement

alpha_aa = "ACDEFGHIKLMNPQRSTVWY"
alpha_nuc = "ATUGCN"

# DNA scoring scheme from Gene 167:GC1
scores_nuc = {"A":{"A": 5,"T":-4,"U":-4,"G":-4,"C":-4,"N":0},
              "T":{"A":-4,"T": 5,"U": 5,"G":-4,"C":-4,"N":0},
              "U":{"A":-4,"T": 5,"U": 5,"G": 5,"C":-4,"N":0},
              "G":{"A":-4,"T":-4,"U":-4,"G": 5,"C":-4,"N":0},
              "C":{"A":-4,"T":-4,"U":-4,"G":-4,"C": 5,"N":0},
              "N":{"A": 0,"T": 0,"U": 0,"G": 0,"C": 0,"N":0}}

# BLOSUM62 matrix from NCBI BLAST
#   (values are scaled and quantized lod scores, c.f. page 59 of the BLAST book)
#     s[i][j] = int(.5+scale*log(q[i][j]/(p[i]*p[j])))
blosum62 = {
    '*':{'*':1,'A':-4,'C':-4,'B':-4,'E':-4,'D':-4,'G':-4,'F':-4,'I':-4,'H':-4,'K':-4,'M':-4,'L':-4,'N':-4,'Q':-4,'P':-4,'S':-4,'R':-4,'T':-4,'W':-4,'V':-4,'Y':-4,'X':-4,'Z':-4},
    'A':{'*':-4,'A':4,'C':0,'B':-2,'E':-1,'D':-2,'G':0,'F':-2,'I':-1,'H':-2,'K':-1,'M':-1,'L':-1,'N':-2,'Q':-1,'P':-1,'S':1,'R':-1,'T':0,'W':-3,'V':0,'Y':-2,'X':-1,'Z':-1},
    'C':{'*':-4,'A':0,'C':9,'B':-3,'E':-4,'D':-3,'G':-3,'F':-2,'I':-1,'H':-3,'K':-3,'M':-1,'L':-1,'N':-3,'Q':-3,'P':-3,'S':-1,'R':-3,'T':-1,'W':-2,'V':-1,'Y':-2,'X':-1,'Z':-3},
    'B':{'*':-4,'A':-2,'C':-3,'B':4,'E':1,'D':4,'G':-1,'F':-3,'I':-3,'H':0,'K':0,'M':-3,'L':-4,'N':3,'Q':0,'P':-2,'S':0,'R':-1,'T':-1,'W':-4,'V':-3,'Y':-3,'X':-1,'Z':1},
    'E':{'*':-4,'A':-1,'C':-4,'B':1,'E':5,'D':2,'G':-2,'F':-3,'I':-3,'H':0,'K':1,'M':-2,'L':-3,'N':0,'Q':2,'P':-1,'S':0,'R':0,'T':-1,'W':-3,'V':-2,'Y':-2,'X':-1,'Z':4},
    'D':{'*':-4,'A':-2,'C':-3,'B':4,'E':2,'D':6,'G':-1,'F':-3,'I':-3,'H':-1,'K':-1,'M':-3,'L':-4,'N':1,'Q':0,'P':-1,'S':0,'R':-2,'T':-1,'W':-4,'V':-3,'Y':-3,'X':-1,'Z':1},
    'G':{'*':-4,'A':0,'C':-3,'B':-1,'E':-2,'D':-1,'G':6,'F':-3,'I':-4,'H':-2,'K':-2,'M':-3,'L':-4,'N':0,'Q':-2,'P':-2,'S':0,'R':-2,'T':-2,'W':-2,'V':-3,'Y':-3,'X':-1,'Z':-2},
    'F':{'*':-4,'A':-2,'C':-2,'B':-3,'E':-3,'D':-3,'G':-3,'F':6,'I':0,'H':-1,'K':-3,'M':0,'L':0,'N':-3,'Q':-3,'P':-4,'S':-2,'R':-3,'T':-2,'W':1,'V':-1,'Y':3,'X':-1,'Z':-3},
    'I':{'*':-4,'A':-1,'C':-1,'B':-3,'E':-3,'D':-3,'G':-4,'F':0,'I':4,'H':-3,'K':-3,'M':1,'L':2,'N':-3,'Q':-3,'P':-3,'S':-2,'R':-3,'T':-1,'W':-3,'V':3,'Y':-1,'X':-1,'Z':-3},
    'H':{'*':-4,'A':-2,'C':-3,'B':0,'E':0,'D':-1,'G':-2,'F':-1,'I':-3,'H':8,'K':-1,'M':-2,'L':-3,'N':1,'Q':0,'P':-2,'S':-1,'R':0,'T':-2,'W':-2,'V':-3,'Y':2,'X':-1,'Z':0},
    'K':{'*':-4,'A':-1,'C':-3,'B':0,'E':1,'D':-1,'G':-2,'F':-3,'I':-3,'H':-1,'K':5,'M':-1,'L':-2,'N':0,'Q':1,'P':-1,'S':0,'R':2,'T':-1,'W':-3,'V':-2,'Y':-2,'X':-1,'Z':1},
    'M':{'*':-4,'A':-1,'C':-1,'B':-3,'E':-2,'D':-3,'G':-3,'F':0,'I':1,'H':-2,'K':-1,'M':5,'L':2,'N':-2,'Q':0,'P':-2,'S':-1,'R':-1,'T':-1,'W':-1,'V':1,'Y':-1,'X':-1,'Z':-1},
    'L':{'*':-4,'A':-1,'C':-1,'B':-4,'E':-3,'D':-4,'G':-4,'F':0,'I':2,'H':-3,'K':-2,'M':2,'L':4,'N':-3,'Q':-2,'P':-3,'S':-2,'R':-2,'T':-1,'W':-2,'V':1,'Y':-1,'X':-1,'Z':-3},
    'N':{'*':-4,'A':-2,'C':-3,'B':3,'E':0,'D':1,'G':0,'F':-3,'I':-3,'H':1,'K':0,'M':-2,'L':-3,'N':6,'Q':0,'P':-2,'S':1,'R':0,'T':0,'W':-4,'V':-3,'Y':-2,'X':-1,'Z':0},
    'Q':{'*':-4,'A':-1,'C':-3,'B':0,'E':2,'D':0,'G':-2,'F':-3,'I':-3,'H':0,'K':1,'M':0,'L':-2,'N':0,'Q':5,'P':-1,'S':0,'R':1,'T':-1,'W':-2,'V':-2,'Y':-1,'X':-1,'Z':3},
    'P':{'*':-4,'A':-1,'C':-3,'B':-2,'E':-1,'D':-1,'G':-2,'F':-4,'I':-3,'H':-2,'K':-1,'M':-2,'L':-3,'N':-2,'Q':-1,'P':7,'S':-1,'R':-2,'T':-1,'W':-4,'V':-2,'Y':-3,'X':-1,'Z':-1},
    'S':{'*':-4,'A':1,'C':-1,'B':0,'E':0,'D':0,'G':0,'F':-2,'I':-2,'H':-1,'K':0,'M':-1,'L':-2,'N':1,'Q':0,'P':-1,'S':4,'R':-1,'T':1,'W':-3,'V':-2,'Y':-2,'X':-1,'Z':0},
    'R':{'*':-4,'A':-1,'C':-3,'B':-1,'E':0,'D':-2,'G':-2,'F':-3,'I':-3,'H':0,'K':2,'M':-1,'L':-2,'N':0,'Q':1,'P':-2,'S':-1,'R':5,'T':-1,'W':-3,'V':-3,'Y':-2,'X':-1,'Z':0},
    'T':{'*':-4,'A':0,'C':-1,'B':-1,'E':-1,'D':-1,'G':-2,'F':-2,'I':-1,'H':-2,'K':-1,'M':-1,'L':-1,'N':0,'Q':-1,'P':-1,'S':1,'R':-1,'T':5,'W':-2,'V':0,'Y':-2,'X':-1,'Z':-1},
    'W':{'*':-4,'A':-3,'C':-2,'B':-4,'E':-3,'D':-4,'G':-2,'F':1,'I':-3,'H':-2,'K':-3,'M':-1,'L':-2,'N':-4,'Q':-2,'P':-4,'S':-3,'R':-3,'T':-2,'W':11,'V':-3,'Y':2,'X':-1,'Z':-3},
    'V':{'*':-4,'A':0,'C':-1,'B':-3,'E':-2,'D':-3,'G':-3,'F':-1,'I':3,'H':-3,'K':-2,'M':1,'L':1,'N':-3,'Q':-2,'P':-2,'S':-2,'R':-3,'T':0,'W':-3,'V':4,'Y':-1,'X':-1,'Z':-2},
    'Y':{'*':-4,'A':-2,'C':-2,'B':-3,'E':-2,'D':-3,'G':-3,'F':3,'I':-1,'H':2,'K':-2,'M':-1,'L':-1,'N':-2,'Q':-1,'P':-3,'S':-2,'R':-2,'T':-2,'W':2,'V':-1,'Y':7,'X':-1,'Z':-2},
    'X':{'*':-4,'A':-1,'C':-1,'B':-1,'E':-1,'D':-1,'G':-1,'F':-1,'I':-1,'H':-1,'K':-1,'M':-1,'L':-1,'N':-1,'Q':-1,'P':-1,'S':-1,'R':-1,'T':-1,'W':-1,'V':-1,'Y':-1,'X':-1,'Z':-1},
    'Z':{'*':-4,'A':-1,'C':-3,'B':1,'E':4,'D':1,'G':-2,'F':-3,'I':-3,'H':0,'K':1,'M':-1,'L':-3,'N':0,'Q':3,'P':-1,'S':0,'R':0,'T':-1,'W':-3,'V':-2,'Y':-2,'X':-1,'Z':4}}

class Dotplot:
    """The result of a windowed dotplot calculation."""
    def __init__(self, hseq, vseq = None, alphabet = alpha_aa,
                 scores = blosum62, W = 25, maxsize = 800**2,
                 hfeatures = [], vfeatures = []):
        """Initialize from sequences, scoring scheme, windowing size,
        and memory limit (a proxy for resolution).  If maxsize is zero,
        the results of the calculation will not be saved (useful for
        debugging and timing).
        """
        self.hseq = hseq
        if(vseq is None):
            vseq = self.hseq
        self.vseq = vseq
        self.hfeatures = hfeatures
        self.vfeatures = vfeatures
        self.rseq = None
        self.alphabet = alphabet
        self.scores = scores

        M = len(self.vseq)
        N = len(self.hseq)

        if(maxsize == 0):
            # Special case: matrix > /dev/null
            self.d = None
            self.update = self.nullupdate

        elif((maxsize is None) or (N*M <= maxsize)):
            # No compression
            self.d = [[0]*N for i in xrange(M)]
            self.update = self.flatupdate
            self.T = 1

        else:
            # Compress to TxT pixels based on maxsize
            self.T = int(sqrt(N*M/float(maxsize)))
            self.hdim = int(ceil(float(N)/self.T))
            self.vdim = int(ceil(float(M)/self.T))
            self.d = [[0]*self.hdim for i in xrange(self.vdim)]
            self.update = self.maxupdate

        self.dotmatrix(N,M,W)

    def nullupdate(self, i, j, v):
        pass

    def flatupdate(self, i, j, v):
        self.d[i][j] = v

    def maxupdate(self, i, j, v):
        m = int(i/self.T)
        n = int(j/self.T)
        self.d[m][n] = max(self.d[m][n],v)
        
    def dotmatrix(self, N,M,W):
        a2i = dict((a,i) for (i,a) in enumerate(self.alphabet))
        hi = [a2i[a] for a in self.hseq]
        vi = [a2i[a] for a in self.vseq]

        zerovec = [0]*N
        newsum = [0]*N
        oldsum = [0]*N

        scorevec = [[self.scores[i][j] for j in self.hseq] for i in self.alphabet]

        W2 = int(W/2)

        for i in xrange(M):
            (oldsum, newsum) = (newsum, oldsum)
            addvec = scorevec[vi[i]]
            if(i < W):
                delvec = zerovec
            else:
                delvec = scorevec[vi[i-W]]

            newsum[0] = addvec[0]
            for j in xrange(1,W):
                newsum[j] = oldsum[j-1]+addvec[j]
            for j in xrange(W,N):
                newsum[j] = oldsum[j-1]+addvec[j]-delvec[j-W]
                if((newsum[j] > 0) and (i >= W)):
                    self.update(i-W2,j-W2,newsum[j]/float(W))

class NucleotideDotplot:
    """A merge of the dotplot calculations for both forward and
    reverse complement orientions of one nucleotide sequence (hseq)
    relative to a second (vseq)."""
    # NucleotideDotplot and Dotplot provide a common interface
    # via their hseq, vseq, d, and T members.  We could make
    # this more explicit by deriving them from a common base
    # class that documented these members as properties.
    def __init__(self, hseq, vseq = None, W = 25, maxsize = 800**2,
                 hfeatures = [], vfeatures = []):
        self.hseq = hseq
        if(vseq is None):
            vseq = self.hseq
        self.vseq = vseq
        self.hfeatures = hfeatures
        self.vfeatures = vfeatures
        self.alphabet = alpha_nuc
        self.scores = scores_nuc
        self.fwd = Dotplot(hseq, vseq, alpha_nuc, scores_nuc, W, maxsize)
        self.rseq = complement(hseq)
        self.rev = Dotplot(self.rseq, vseq, alpha_nuc, scores_nuc,
                           W, maxsize)
        t = [list(reversed(i)) for i in self.rev.d]
        self.rev.d = t
        self.d = [i[:] for i in self.rev.d]
        for i in xrange(len(self.d)):
            for j in xrange(len(self.d[0])):
                self.d[i][j] = max(self.d[i][j], self.fwd.d[i][j])

        self.T = self.fwd.T
