"""Qt4-specific dotplot GUI elements."""

from matplotlib.backends import qt4_compat
use_pyside = qt4_compat.QT_API == qt4_compat.QT_API_PYSIDE

if use_pyside:
    from PySide import QtGui, QtCore
else:
    from PyQt4 import QtGui, QtCore

class QtAlignmentFrame(QtGui.QWidget):
    """A frame displaying the ungapped alignment corresponding to the
    cursor position in the dotplot."""
    def __init__(self, parent = None, hseq = "", vseq = "",
                 hpos = 62, vpos = 62, pad = 62, scores = None):
        """Initialize from sequences and starting position.
        If scores is given, it will be used for similarity coloring."""
        QtGui.QWidget.__init__(self, parent)
        self.scores = scores
        self.hseq = hseq
        self.vseq = vseq
        self.hpos = hpos
        self.vpos = vpos
        self.pad = pad

        self.setWindowTitle('Alignment')    
        self.text = QtGui.QLabel("",self)
        #self.text.setTextFormat(1)

        self.update_text()

        self.text.adjustSize()

        self.show()
        
    def update_text(self, h_id = None):
        if(h_id == None):
            h_id = self.hpos+1
        
        header = " "*self.pad+str(h_id)+"\n"+" "*self.pad+"|\n"
        footer = " "*self.pad+"|\n"+" "*self.pad+str(self.vpos+1)

        aseqs = []
        for (pos, seq) in ((self.hpos, self.hseq), (self.vpos, self.vseq)):
            left = pos - self.pad
            right = min(len(self.hseq), pos+self.pad)+1
            if(left < 0):
                s = " "*(-left)+seq[:right]
            else:
                s = seq[left:right]
            aseqs.append(s)

        ident_color = "cyan"
        sim_color = "#56b2de"

        colors = []
        lastcolor = None
        htext = vtext = ""
        for (i, (h,v)) in enumerate(zip(aseqs[0],aseqs[1])):
            try:
                if(h == v):
                    if(h != ' '):
                        curcolor = ident_color
                elif(self.scores[h][v] > 0):
                    curcolor = sim_color
                else:
                    curcolor = None
            except KeyError:
                # handle gap comparisons as exceptions
                curcolor = None

            if(curcolor != lastcolor):
                if(lastcolor is not None):
                    htext += "</SPAN>"
                    vtext += "</SPAN>"
                if(curcolor is not None):
                    htext += ("<SPAN style='background-color: %s'>" % curcolor)
                    vtext += ("<SPAN style='background-color: %s'>" % curcolor)
            htext += h
            vtext += v

            lastcolor = curcolor

        if(lastcolor is not None):
            htext += "</SPAN>"
            vtext += "</SPAN>"

        htext += "\n"
        vtext += "\n"
        
        self.text.setText("<PRE>"+header+htext+vtext+footer+"</PRE>")
        self.plaintext = header+aseqs[0]+"\n"+aseqs[1]+"\n"+footer+"\n"

import matplotlib.pyplot as plt

class QtDotMediator:
    def __init__(self, dotplot):
        import numpy as np
        
        # Matplotlib (generic)
        (self.fig, self.ax) = plt.subplots(1,1)
        self.dotplot = dotplot
        self.D = np.array(dotplot.d)
        plt.imshow(self.D, interpolation = "none", aspect="auto", cmap = "Greys",
                   extent = [1,len(dotplot.hseq), len(dotplot.vseq), 1])
        self.vpos = self.hpos = 62
        self.vp_line = self.ax.axhline(self.vpos, color = "blue")
        self.hp_line = self.ax.axvline(self.hpos, color = "blue")
        
        # Qt-specific
        #  This will probably need the create_window wrapper
        
        self.alignment = QtAlignmentFrame(None, dotplot.hseq, dotplot.vseq,
                                          vpos = self.vpos, hpos = self.hpos,
                                          scores = dotplot.scores)
        self.update()
        
        # Disable default plot keybindings
        key_callbacks = self.fig.canvas.callbacks.callbacks[
            "key_press_event"].keys()
        for i in key_callbacks:
            self.fig.canvas.mpl_disconnect(i)
        # Substitute our own keybindings
        self.fig.canvas.mpl_connect("key_press_event", self.keypress)
        # For mouse clicks, we'll do our cross-hairs update in series
        # with the default mouse interactions
        self.fig.canvas.mpl_connect("button_press_event", self.click)

    def update(self):
        self.vp_line.set_ydata([self.vpos, self.vpos])
        self.hp_line.set_xdata([self.hpos, self.hpos])
        plt.draw()
        
        self.alignment.vpos = self.vpos
        self.alignment.hpos = self.hpos
        self.alignment.update_text()
        
    def click(self, event):
        (x,y) = map(int,(event.xdata, event.ydata))
        if(y < 0):
            self.vpos = 0
        elif(y >= len(self.dotplot.vseq)):
            self.vpos = len(self.dotplot.vseq)-1
        else:
            self.vpos = y
        if(x < 0):
            self.hpos = 0
        elif(x >= len(self.dotplot.hseq)):
            self.hpos = len(self.dotplot.hseq)-1
        else:
            self.hpos = x
        self.update()
        
    def keypress(self, event):
        if(event.key == "enter"):
            self.print_alignment()
        elif(event.key in ("up","k")):
            self.vleft()
        elif(event.key in ("down","j")):
            self.vright()
        elif(event.key in ("left","h")):
            self.hleft()
        elif(event.key in ("right","l")):
            self.hright()
        elif(event.key == "b"):
            self.offleft()
        elif(event.key == "n"):
            self.onright()
        elif(event.key == "y"):
            self.onleft()
        elif(event.key == "u"):
            self.offright()
        
    def vleft(self):
        self.vpos = max(0, self.vpos-1)
        self.update()
    def vright(self):
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def hleft(self):
        self.hpos = max(0, self.hpos-1)
        self.update()
    def hright(self):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.update()
    def onleft(self):
        self.hpos = max(0, self.hpos-1)
        self.vpos = max(0, self.vpos-1)
        self.update()
    def onright(self):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def offleft(self):
        self.hpos = max(0, self.hpos-1)
        self.vpos = min(len(self.dotplot.vseq), self.vpos+1)
        self.update()
    def offright(self):
        self.hpos = min(len(self.dotplot.hseq), self.hpos+1)
        self.vpos = max(0, self.vpos-1)
        self.update()

    def print_alignment(self):
        print self.alignment.plaintext
